{-# LANGUAGE QuasiQuotes #-}

{-
A very simple worker to test Database.PGMQ.Worker.
-}
    
module Main
where

import Control.Applicative ((<|>))
import Control.Concurrent (forkIO, threadDelay)
import Control.Exception (Exception, throwIO)
import Data.Aeson (FromJSON(..), ToJSON(..), object, (.=), (.:), withObject, withText)
import Data.Text qualified as T
import Database.PostgreSQL.Simple qualified as PSQL
import Database.PGMQ.Simple qualified as PGMQ
import Database.PGMQ.Types qualified as PGMQ
import Database.PGMQ.Worker qualified as PGMQW


data SimpleException = SimpleException String
  deriving Show
instance Exception SimpleException
    

data Message =
    Ping
  | Echo String
  | Wait Int
  | Error String
  deriving (Eq, Show)
instance ToJSON Message where
  toJSON Ping = toJSON ("ping" :: String)
  toJSON (Echo s) = toJSON $ object [ ("echo" .= s) ]
  toJSON (Wait s) = toJSON $ object [ ("wait" .= s) ]
  toJSON (Error e) = toJSON $ object [ ("error" .= e) ]
instance FromJSON Message where
  parseJSON v = parsePing v <|> parseEcho v <|> parseWait v <|> parseError v
    where
      parsePing = withText "Message (Ping)" $ \s -> do
        case s of
          "ping" -> pure Ping
          s'     -> fail $ T.unpack s'
      parseEcho = withObject "Message (Echo)" $ \o -> do
        s <- o .: "echo"
        return $ Echo s
      parseWait = withObject "Message (Wait)" $ \o -> do
        s <- o .: "wait"
        return $ Wait s
      parseError = withObject "Message (Error)" $ \o -> do
        e <- o .: "error"
        return $ Error e

    
performAction :: PGMQW.State Message -> PGMQW.JobMessage Message -> IO ()
performAction s (PGMQ.Message { message = job }) = do
  case PGMQW.getJob job of
    Ping -> putStrLn $ PGMQW.formatStr s "ping"
    Echo str -> putStrLn $ PGMQW.formatStr s ("echo " <> str)
    Wait int -> do 
      putStrLn $ PGMQW.formatStr s ( "waiting " <> show int <> " seconds")
      threadDelay (int*second)
    Error err -> do
      putStrLn $ PGMQW.formatStr s ("error " <> err)
      throwIO $ SimpleException $ "Error " <> err

second :: Int
second = 1000000
    
main :: IO ()
main = do
  let connInfo = PSQL.defaultConnectInfo { PSQL.connectUser = "postgres"
                                         , PSQL.connectDatabase = "postgres" }

  let queue = "simple_worker"
  
  -- let workersLst = [1, 2, 3, 4] :: [Int]
  let workersLst = [1, 2] :: [Int]
  -- let tasksLst = [101, 102, 102, 103, 104, 105, 106, 107] :: [Int]
  let tasksLst = [101] :: [Int]
  -- let tasksLst = [] :: [Int]

  mapM_ (\idx -> do
      let state = PGMQW.newState connInfo ("worker " <> show idx) performAction queue
      forkIO $ PGMQW.run state
    ) workersLst
  
  -- delay so that the worker can initialize and settle
  threadDelay second

  conn <- PSQL.connect connInfo

  -- SendJob wrapper
  let mkJob msg = PGMQW.mkDefaultSendJob conn queue msg
    
  mapM_ (\idx -> do
    PGMQW.sendJob $ mkJob $ Ping
    PGMQW.sendJob $ mkJob $ Wait 1
    PGMQW.sendJob $ mkJob $ Echo $ "hello " <> show idx
    PGMQW.sendJob $ mkJob $ Error $ "error " <> show idx
    ) tasksLst

  -- a job that will timeout
  let timedOut = (mkJob (Wait 5)) { PGMQW.vt = Just 1
                                  , PGMQW.toStrat = PGMQW.TSRepeatNElseArchive 3 }
  PGMQW.sendJob timedOut
    
  threadDelay (10*second)

  metrics <- PGMQ.getMetrics conn queue
  putStrLn $ "metrics: " <> show metrics

  return ()

